"""Test deltacycle._kernel.finish"""

import logging

from pytest import LogCaptureFixture

from deltacycle import Kernel, create_task, finish, get_kernel, irun, run, sleep

logger = logging.getLogger("deltacycle")


async def ctl():
    logger.info("enter")
    await sleep(100)

    # Force all PING threads to stop immediately
    logger.info("finish")
    finish()


async def ping(period: int):
    logger.info("enter")
    while True:
        await sleep(period)
        logger.info("PING")


EXP1 = {
    # CTL
    (0, "CTL", "enter"),
    (100, "CTL", "finish"),
    # FOO
    (0, "FOO", "enter"),
    (3, "FOO", "PING"),
    (6, "FOO", "PING"),
    (9, "FOO", "PING"),
    (12, "FOO", "PING"),
    (15, "FOO", "PING"),
    (18, "FOO", "PING"),
    (21, "FOO", "PING"),
    (24, "FOO", "PING"),
    (27, "FOO", "PING"),
    (30, "FOO", "PING"),
    (33, "FOO", "PING"),
    (36, "FOO", "PING"),
    (39, "FOO", "PING"),
    (42, "FOO", "PING"),
    (45, "FOO", "PING"),
    (48, "FOO", "PING"),
    (51, "FOO", "PING"),
    (54, "FOO", "PING"),
    (57, "FOO", "PING"),
    (60, "FOO", "PING"),
    (63, "FOO", "PING"),
    (66, "FOO", "PING"),
    (69, "FOO", "PING"),
    (72, "FOO", "PING"),
    (75, "FOO", "PING"),
    (78, "FOO", "PING"),
    (81, "FOO", "PING"),
    (84, "FOO", "PING"),
    (87, "FOO", "PING"),
    (90, "FOO", "PING"),
    (93, "FOO", "PING"),
    (96, "FOO", "PING"),
    (99, "FOO", "PING"),
    # BAR
    (0, "BAR", "enter"),
    (5, "BAR", "PING"),
    (10, "BAR", "PING"),
    (15, "BAR", "PING"),
    (20, "BAR", "PING"),
    (25, "BAR", "PING"),
    (30, "BAR", "PING"),
    (35, "BAR", "PING"),
    (40, "BAR", "PING"),
    (45, "BAR", "PING"),
    (50, "BAR", "PING"),
    (55, "BAR", "PING"),
    (60, "BAR", "PING"),
    (65, "BAR", "PING"),
    (70, "BAR", "PING"),
    (75, "BAR", "PING"),
    (80, "BAR", "PING"),
    (85, "BAR", "PING"),
    (90, "BAR", "PING"),
    (95, "BAR", "PING"),
    # FIZ
    (0, "FIZ", "enter"),
    (7, "FIZ", "PING"),
    (14, "FIZ", "PING"),
    (21, "FIZ", "PING"),
    (28, "FIZ", "PING"),
    (35, "FIZ", "PING"),
    (42, "FIZ", "PING"),
    (49, "FIZ", "PING"),
    (56, "FIZ", "PING"),
    (63, "FIZ", "PING"),
    (70, "FIZ", "PING"),
    (77, "FIZ", "PING"),
    (84, "FIZ", "PING"),
    (91, "FIZ", "PING"),
    (98, "FIZ", "PING"),
    # BUZ
    (0, "BUZ", "enter"),
    (11, "BUZ", "PING"),
    (22, "BUZ", "PING"),
    (33, "BUZ", "PING"),
    (44, "BUZ", "PING"),
    (55, "BUZ", "PING"),
    (66, "BUZ", "PING"),
    (77, "BUZ", "PING"),
    (88, "BUZ", "PING"),
    (99, "BUZ", "PING"),
}


def test_finish1(caplog: LogCaptureFixture):
    caplog.set_level(logging.INFO, logger="deltacycle")

    async def main():
        create_task(ctl(), name="CTL")
        create_task(ping(3), name="FOO")
        create_task(ping(5), name="BAR")
        create_task(ping(7), name="FIZ")
        create_task(ping(11), name="BUZ")

    # Subsequent calls to run() have no effect
    run(main())

    kernel = get_kernel()
    assert kernel is not None
    assert kernel.state() is Kernel.State.FINISHED

    msgs = {(r.time, r.taskName, r.getMessage()) for r in caplog.records}
    assert msgs == EXP1


def test_finish2(caplog: LogCaptureFixture):
    caplog.set_level(logging.INFO, logger="deltacycle")

    async def main():
        create_task(ctl(), name="CTL")
        create_task(ping(3), name="FOO")
        create_task(ping(5), name="BAR")
        create_task(ping(7), name="FIZ")
        create_task(ping(11), name="BUZ")

    # Subsequent calls to run() have no effect
    list(irun(main()))

    kernel = get_kernel()
    assert kernel is not None
    assert kernel.state() is Kernel.State.FINISHED
    assert kernel.done()

    msgs = {(r.time, r.taskName, r.getMessage()) for r in caplog.records}
    assert msgs == EXP1



"""Execution Kernel"""

import logging
from collections import defaultdict
from collections.abc import Generator
from enum import IntEnum

from ._event import Event
from ._task import PendQueue, Signal, Task, TaskCoro
from ._variable import Variable

# Awaitables
type AW = Task | Event | Variable
type CallValue = tuple[Task.Command] | tuple[Task.Command, AW | Signal]

logger = logging.getLogger("deltacycle")


class _SuspendResume:
    """Suspend/Resume current task.

    Use case:
    1. Current task A suspends itself: RUNNING => WAITING
    2. Kernel chooses PENDING tasks ..., T
    3. ... Task T wakes up task A w/ value X: WAITING => PENDING
    4. Kernel chooses PENDING tasks ..., A: PENDING => RUNNING
    5. Task A resumes with value X

    The value X can be used to pass information to the task.
    """

    def __await__(self) -> Generator[None, AW | None, AW | None]:
        # Suspend
        value = yield
        # Resume
        return value


class _Finish(Exception):
    """Force the simulation to stop."""


class Kernel:
    """Simulation Kernel.

    Responsible for:

    * Scheduling and executing all tasks
    * Updating all model state

    This is a low level API.
    User code is not expected to interact with it directly.
    To run a simulation, use the run and irun functions.
    """

    _index = 0

    class State(IntEnum):
        """
        Transitions::

            INIT -> RUNNING -> COMPLETED
                            -> FINISHED
        """

        # Initialized
        INIT = 0b001

        # Currently running
        RUNNING = 0b010

        # All tasks completed
        COMPLETED = 0b100

        # finish() called
        FINISHED = 0b101

    _done = State.COMPLETED & State.FINISHED

    _state_transitions = {
        State.INIT: {
            State.RUNNING,
        },
        State.RUNNING: {
            State.COMPLETED,
            State.FINISHED,
        },
    }

    init_time = -1
    start_time = 0

    main_name = "main"
    main_priority = 0

    def __init__(self):
        self._name = f"Kernel-{self.__class__._index}"
        self.__class__._index += 1

        self._state = self.State.INIT

        # Simulation time
        self._time: int = self.init_time

        # Main task
        self._main: Task | None = None

        # Currently executing task
        self._task: Task | None = None
        self._task_index = 0

        # Task queue
        self._queue = PendQueue()

        # Task => Event mapping table
        self._task2events: defaultdict[Task, set[Event]] = defaultdict(set)

        # Task => Variable mapping table
        self._task2vars: defaultdict[Task, set[Variable]] = defaultdict(set)

        # Model variables
        self._touched: set[Variable] = set()

    def _set_state(self, state: State):
        assert state in self._state_transitions[self._state]
        logger.debug("%s: %s => %s", self._name, self._state.name, state.name)
        self._state = state

    def state(self) -> State:
        """Current simulation state."""
        return self._state

    def time(self) -> int:
        """Current simulation time."""
        return self._time

    @property
    def main(self) -> Task:
        """Parent task of all other tasks."""
        assert self._main is not None
        return self._main

    def task(self) -> Task:
        """Currently running task."""
        assert self._task is not None
        return self._task

    def done(self) -> bool:
        return bool(self._state & self._done)

    # Scheduling methods
    def call_soon(self, task: Task, value: CallValue):
        self._queue.push((self._time, task, value))

    def call_later(self, delay: int, task: Task, value: CallValue):
        self._queue.push((self._time + delay, task, value))

    def call_at(self, when: int, task: Task, value: CallValue):
        self._queue.push((when, task, value))

    def create_main(self, coro: TaskCoro):
        assert self._time == self.init_time
        self._main = Task(coro, self.main_name, self.main_priority)
        self.call_at(self.start_time, self._main, value=(Task.Command.START,))

    def create_task(
        self,
        coro: TaskCoro,
        name: str | None = None,
        priority: int = 0,
    ) -> Task:
        assert self._time >= self.start_time
        if name is None:
            name = f"Task-{self._task_index}"
            self._task_index += 1
        task = Task(coro, name, priority)
        self.call_soon(task, value=(Task.Command.START,))
        return task

    async def switch_coro(self) -> AW | None:
        assert self._task is not None
        # Suspend
        self._task._set_state(Task.State.PENDING)
        value = await _SuspendResume()
        # Resume
        return value

    def switch_gen(self) -> Generator[None, AW, AW]:
        assert self._task is not None
        # Suspend
        self._task._set_state(Task.State.PENDING)
        value = yield
        # Resume
        return value

    def touch(self, v: Variable):
        self._touched.add(v)

    def _update(self):
        while self._touched:
            v = self._touched.pop()
            v.update()

    def _start(self):
        if self._state is self.State.INIT:
            self._set_state(self.State.RUNNING)
        elif self._state is not self.State.RUNNING:
            s = f"Kernel has invalid state: {self._state.name}"
            raise RuntimeError(s)

    def _iter_time_slot(self, time: int) -> Generator[tuple[Task, CallValue], None, None]:
        """Iterate through all tasks in a time slot.

        The first task has already been peeked.
        This is a do-while loop.
        """
        task, value = self._queue.pop()
        yield (task, value)
        while self._queue and self._queue.peek() == time:
            task, value = self._queue.pop()
            yield (task, value)

    def _finish(self):
        self._queue.clear()
        self._task2events.clear()
        self._task2vars.clear()
        self._touched.clear()
        self._set_state(self.State.FINISHED)

    def _call(self, limit: int | None):
        self._start()

        while self._queue:
            # Peek when next event is scheduled
            time = self._queue.peek()

            # Protect against time traveling tasks
            assert time > self._time

            # Halt if we hit the run limit
            if limit is not None and time >= limit:
                return

            # Otherwise, advance to new timeslot
            self._time = time

            # Execute time slot
            for task, args in self._iter_time_slot(time):
                self._task = task
                try:
                    task._do_run(args)
                except _Finish:
                    self._finish()
                    return
                except StopIteration as exc:
                    task._do_result(exc)
                except Exception as exc:
                    task._do_except(exc)

            # Update simulation state
            self._update()

        # All tasks exhausted
        self._set_state(self.State.COMPLETED)

    def _iter(self) -> Generator[int, None, None]:
        self._start()

        while self._queue:
            # Peek when next event is scheduled
            time = self._queue.peek()

            # Protect against time traveling tasks
            assert time > self._time

            # Yield before entering new timeslot
            yield time

            # Advance to new timeslot
            self._time = time

            # Execute time slot
            for task, args in self._iter_time_slot(time):
                self._task = task
                try:
                    task._do_run(args)
                except _Finish:
                    self._finish()
                    return
                except StopIteration as exc:
                    task._do_result(exc)
                except Exception as exc:
                    task._do_except(exc)

            # Update simulation state
            self._update()

        # All tasks exhausted
        self._set_state(self.State.COMPLETED)

    def __call__(self, ticks: int | None = None, until: int | None = None):
        # Determine the run limit
        match ticks, until:
            # Run until no tasks left
            case None, None:
                limit = None
            # Run until an absolute time
            case None, int():
                limit = until
            # Run until a number of ticks in the future
            case int(), None:
                limit = max(self.start_time, self._time) + ticks
            case int(), int():
                limit = max(self.start_time, self._time) + ticks
                limit = min(limit, until)
            case _:
                s = "Expected either ticks or until to be int | None"
                raise TypeError(s)

        self._call(limit)

    def __iter__(self) -> Generator[int, None, None]:
        yield from self._iter()


def finish():
    """Halt all incomplete coroutines, and immediately exit simulation.

    Clear all kernel data, and transition state to FINISHED.
    """
    raise _Finish()



"""Test deltacycle.Semaphore"""

import logging

import pytest
from pytest import LogCaptureFixture

from deltacycle import BoundedSemaphore, Semaphore, create_task, run, sleep

logger = logging.getLogger("deltacycle")


async def use_acquire_release(sem: Semaphore, t1: int, t2: int):
    logger.info("enter")

    await sleep(t1)

    logger.info("attempt get")
    await sem.get()
    logger.info("acquired")

    try:
        await sleep(t2)
    finally:
        logger.info("put")
        sem.put()

    await sleep(10)
    logger.info("exit")


async def use_with(sem: Semaphore, t1: int, t2: int):
    logger.info("enter")

    await sleep(t1)

    logger.info("attempt get")
    async with sem:
        logger.info("acquired")
        await sleep(t2)
    logger.info("put")

    await sleep(10)
    logger.info("exit")


EXP = {
    # 0
    (0, "0", "enter"),
    (10, "0", "attempt get"),
    (10, "0", "acquired"),
    (20, "0", "put"),
    (30, "0", "exit"),
    # 1
    (0, "1", "enter"),
    (11, "1", "attempt get"),
    (11, "1", "acquired"),
    (21, "1", "put"),
    # 2
    (0, "2", "enter"),
    (12, "2", "attempt get"),
    (12, "2", "acquired"),
    (22, "2", "put"),
    (32, "2", "exit"),
    # 3
    (0, "3", "enter"),
    (13, "3", "attempt get"),
    (13, "3", "acquired"),
    (23, "3", "put"),
    (33, "3", "exit"),
    # 4
    (0, "4", "enter"),
    (14, "4", "attempt get"),
    (20, "4", "acquired"),
    (30, "4", "put"),
    (40, "4", "exit"),
    # 5
    (0, "5", "enter"),
    (15, "5", "attempt get"),
    (21, "5", "acquired"),
    (31, "5", "put"),
    (41, "5", "exit"),
    # 6
    (0, "6", "enter"),
    (16, "6", "attempt get"),
    (22, "6", "acquired"),
    (32, "6", "put"),
    (42, "6", "exit"),
    # 7
    (0, "7", "enter"),
    (17, "7", "attempt get"),
    (23, "7", "acquired"),
    (31, "1", "exit"),
    (33, "7", "put"),
    (43, "7", "exit"),
}


def test_acquire_release(caplog: LogCaptureFixture):
    caplog.set_level(logging.INFO, logger="deltacycle")

    async def main():
        sem = Semaphore(4)
        for i in range(8):
            create_task(use_acquire_release(sem, i + 10, 10), name=f"{i}")

    run(main())

    msgs = {(r.time, r.taskName, r.getMessage()) for r in caplog.records}
    assert msgs == EXP


def test_async_with(caplog: LogCaptureFixture):
    caplog.set_level(logging.INFO, logger="deltacycle")

    async def main():
        sem = Semaphore(4)
        for i in range(8):
            create_task(use_with(sem, i + 10, 10), name=f"{i}")

    run(main())

    msgs = {(r.time, r.taskName, r.getMessage()) for r in caplog.records}
    assert msgs == EXP


def test_unbounded():
    async def use_unbounded():
        sem = Semaphore(2)

        await sem.get()
        await sem.get()
        sem.put()
        sem.put()

        # No exception!
        sem.put()

    run(use_unbounded())


def test_bounded():
    async def use_bounded():
        sem = BoundedSemaphore(2)

        await sem.get()
        await sem.get()

        sem.put()
        sem.put()

        # Exception!
        with pytest.raises(ValueError):
            sem.put()

    async def main():
        create_task(use_bounded())

    run(main())


def test_init_bad_values():
    with pytest.raises(ValueError):
        _ = Semaphore(0)

    with pytest.raises(ValueError):
        _ = Semaphore(-1)



"""Top-level functions."""

from collections.abc import Generator
from typing import Any

from ._event import Event
from ._kernel import Kernel
from ._task import Predicate, Task, TaskCoro
from ._variable import Variable

_kernel: Kernel | None = None


def get_running_kernel() -> Kernel:
    """Return currently running kernel.

    Returns:
        Kernel instance

    Raises:
        RuntimeError: No kernel, or kernel is not currently running.
    """
    if _kernel is None:
        raise RuntimeError("No kernel")
    if _kernel.state() is not Kernel.State.RUNNING:
        raise RuntimeError("Kernel not RUNNING")
    return _kernel


def get_kernel() -> Kernel | None:
    """Get the current kernel."""
    return _kernel


def set_kernel(kernel: Kernel | None = None):
    """Set the current kernel."""
    global _kernel  # noqa: PLW0603
    _kernel = kernel


def _get_kt() -> tuple[Kernel, Task]:
    kernel = get_running_kernel()
    return kernel, kernel.task()


def get_current_task() -> Task:
    """Return currently running task.

    Returns:
        Task instance

    Raises:
        RuntimeError: No kernel, or kernel is not currently running.
    """
    _, task = _get_kt()
    return task


def create_task(
    coro: TaskCoro,
    name: str | None = None,
    priority: int = 0,
) -> Task:
    """Create a task, and schedule it to start soon."""
    kernel = get_running_kernel()
    return kernel.create_task(coro, name, priority)


def now() -> int:
    """Return current simulation time.

    Returns:
        int time

    Raises:
        RuntimeError: No kernel, or kernel is not currently running.
    """
    kernel = get_running_kernel()
    return kernel.time()


def _run_pre(coro: TaskCoro | None, kernel: Kernel | None) -> Kernel:
    if kernel is None:
        kernel = Kernel()
        set_kernel(kernel)
        if coro is None:
            raise ValueError("New kernel requires a valid coro arg")
        assert coro is not None
        kernel.create_main(coro)
    else:
        set_kernel(kernel)

    return kernel


def run(
    coro: TaskCoro | None = None,
    kernel: Kernel | None = None,
    ticks: int | None = None,
    until: int | None = None,
) -> Any:
    """Run a simulation.

    If a simulation hits the run limit, it will exit and return None.
    That simulation may be resumed any number of times.
    If all tasks are exhausted, return the main coroutine result.

    Args:
        coro: Optional main coroutine.
            Required if creating a new kernel.
            Ignored if using an existing kernel.
        kernel: Optional Kernel instance.
            If not provided, a new kernel will be created.
        ticks: Optional relative run limit.
            If provided, run for *ticks* simulation time steps.
        until: Optional absolute run limit.
            If provided, run until *ticks* simulation time steps.

    Returns:
        If the main coroutine runs til completion, return its result.
        Otherwise, return ``None``.

    Raises:
        ValueError: Creating a new kernel, but no coro provided.
        TypeError: ticks and until args conflict.
        RuntimeError: The kernel is in an invalid state.
    """
    kernel = _run_pre(coro, kernel)
    kernel(ticks, until)

    if kernel.main.done():
        return kernel.main.result()


def irun(
    coro: TaskCoro | None = None,
    kernel: Kernel | None = None,
) -> Generator[int, None, Any]:
    """Iterate a simulation.

    Iterated simulations do not have a run limit.
    It is the user's responsibility to break at the desired time.
    If all tasks are exhausted, return the main coroutine result.

    Args:
        coro: Optional main coroutine.
            Required if creating a new kernel.
            Ignored if using an existing kernel.
        kernel: Optional Kernel instance.
            If not provided, a new kernel will be created.

    Yields:
        int time immediately *before* the next time slot executes.

    Returns:
        main coroutine result.

    Raises:
        ValueError: Creating a new kernel, but no coro provided.
        TypeError: ticks and until args conflict.
        RuntimeError: The kernel is in an invalid state.
    """
    kernel = _run_pre(coro, kernel)
    yield from kernel

    assert kernel.main.done()
    return kernel.main.result()


async def sleep(delay: int):
    """Suspend the task, and wake up after a delay."""
    kernel = get_running_kernel()
    task = kernel.task()
    kernel.call_later(delay, task, value=(Task.Command.RESUME,))
    y = await kernel.switch_coro()
    assert y is None


async def any_event(*events: Event) -> Event:
    """Resume execution after first event.

    Suspend execution of the current task;
    Resume after any event in the sensitivity list,

    Args:
        events: Tuple of Event, a sensitivity list.

    Returns:
        The Event instance that triggered the task to resume.
    """
    kernel, task = _get_kt()

    # Search for first set event
    fst = None
    for e in events:
        if e:
            fst = e
            break

    # No events set yet
    if fst is None:
        # Await first event to be set
        for e in events:
            e._wait(task)
        fst = await kernel.switch_coro()
        assert isinstance(fst, Event)

    return fst


async def any_var(vps: dict[Variable, Predicate]) -> Variable:
    """Resume execution upon predicated variable change.

    Suspend execution of the current task;
    Resume when any variable in the sensitivity list changes,
    *and* the predicate function evaluates to True.
    If the predicate function is None, it will default to *any* change.

    Args:
        vps: Dict of Variable => Predicate mappings, a sensitivity list.

    Returns:
        The Variable instance that triggered the task to resume.
    """
    kernel, task = _get_kt()

    for v, p in vps.items():
        v._wait(p, task)
    v = await kernel.switch_coro()
    assert isinstance(v, Variable)
    return v



"""Delta Cycle

Credit to David Beazley's "Build Your Own Async" tutorial for inspiration:
https://www.youtube.com/watch?v=Y4Gt3Xjd7G8
"""

import logging
from logging import Filter, LogRecord

from ._event import Event, EventList
from ._kernel import Kernel, finish
from ._queue import Queue
from ._semaphore import BoundedSemaphore, Lock, Semaphore
from ._task import (
    Interrupt,
    Predicate,
    Signal,
    Task,
    TaskCoro,
    TaskGroup,
)
from ._top import (
    any_event,
    any_var,
    create_task,
    get_current_task,
    get_kernel,
    get_running_kernel,
    irun,
    now,
    run,
    set_kernel,
    sleep,
)
from ._variable import Aggregate, AggrItem, AggrValue, Singular, Value, Variable

# Customize logging
logger = logging.getLogger(__name__)


class DeltaCycleFilter(Filter):
    def filter(self, record: LogRecord) -> bool:
        try:
            kernel = get_running_kernel()
        except RuntimeError:
            record.time = -1
            record.taskName = None
        else:
            record.time = kernel.time()
            record.taskName = kernel.task().name
        return True


logger.addFilter(DeltaCycleFilter())


__all__ = [
    # kernel
    "Kernel",
    "finish",
    "get_running_kernel",
    "get_kernel",
    "set_kernel",
    "run",
    "irun",
    "now",
    "sleep",
    # event
    "Event",
    "EventList",
    "any_event",
    # queue
    "Queue",
    # semaphore
    "Semaphore",
    "BoundedSemaphore",
    "Lock",
    # task
    "Predicate",
    "TaskCoro",
    "Signal",
    "Interrupt",
    "Task",
    "TaskGroup",
    "create_task",
    "get_current_task",
    # variable
    "Variable",
    "Value",
    "Singular",
    "Aggregate",
    "AggrItem",
    "AggrValue",
    "any_var",
]



"""Semaphore synchronization primitive"""

from types import TracebackType
from typing import override

from ._kernel_if import KernelIf
from ._task import Task, WaitFifo


class Semaphore(KernelIf):
    """Semaphore to synchronize tasks.

    Permits number of put() > resource count.
    """

    def __init__(self, value: int = 1):
        if value < 1:
            raise ValueError(f"Expected value >= 1, got {value}")
        self._cnt = value
        self._waiting = WaitFifo()

    async def __aenter__(self):
        await self.get()
        return self

    async def __aexit__(
        self,
        exc_type: type[Exception],
        exc: Exception,
        traceback: TracebackType,
    ):
        self.put()

    def put(self):
        assert self._cnt >= 0
        if self._waiting:
            task = self._waiting.pop()
            self._kernel.call_soon(task, value=(Task.Command.RESUME,))
        else:
            self._cnt += 1

    def try_get(self) -> bool:
        assert self._cnt >= 0
        if self._cnt == 0:
            return False
        self._cnt -= 1
        return True

    async def get(self):
        assert self._cnt >= 0
        if self._cnt == 0:
            task = self._kernel.task()
            self._waiting.push(task)
            y = await self._kernel.switch_coro()
            assert y is None
        else:
            self._cnt -= 1


class BoundedSemaphore(Semaphore):
    """Bounded Semaphore to synchronize tasks.

    Like Semaphore, but raises ValueError when
    number of put() > resource count.
    """

    def __init__(self, value: int = 1):
        super().__init__(value)
        self._maxcnt = value

    @override
    def put(self):
        assert self._cnt >= 0
        if self._waiting:
            task = self._waiting.pop()
            self._kernel.call_soon(task, value=(Task.Command.RESUME,))
        else:
            if self._cnt == self._maxcnt:
                raise ValueError("Cannot put")
            self._cnt += 1


class Lock(BoundedSemaphore):
    """Mutex lock to synchronize tasks."""

    def __init__(self):
        super().__init__(value=1)



"""Test bugs"""

import logging

from pytest import LogCaptureFixture

from deltacycle import TaskGroup, run, sleep

from .common import Bool

logger = logging.getLogger("deltacycle")


EXP2 = {
    (5, "do_stuff", "first"),
    (15, "do_stuff", "second"),
    (25, "do_stuff", "third"),
    (35, "do_stuff", "fourth"),
}


def test_2(caplog: LogCaptureFixture):
    caplog.set_level(logging.INFO, logger="deltacycle")

    clock = Bool(name="clock")

    async def do_stuff():
        await clock.posedge()
        logger.info("first")
        await clock.posedge()
        logger.info("second")
        await clock.posedge()
        logger.info("third")
        await clock.posedge()
        logger.info("fourth")

    async def drv_clock():
        clock.next = False
        while True:
            await sleep(5)
            clock.next = not clock.value

    async def main():
        async with TaskGroup() as tg:
            tg.create_task(drv_clock(), name="drv_clock")
            tg.create_task(do_stuff(), name="do_stuff")

    run(main(), until=100)

    msgs = {(r.time, r.taskName, r.getMessage()) for r in caplog.records}
    assert msgs == EXP2



"""Simulate a 4-bit adder."""

import logging

from pytest import LogCaptureFixture

from deltacycle import any_var, create_task, run, sleep

from .common import Bool

logger = logging.getLogger("deltacycle")


# a, b, ci, s, co
VALS = [
    (False, False, False, False, False),
    (False, False, True, True, False),
    (False, True, False, True, False),
    (False, True, True, False, True),
    (True, False, False, True, False),
    (True, False, True, False, True),
    (True, True, False, False, True),
    (True, True, True, True, True),
]


def test_add(caplog: LogCaptureFixture):
    """Test 4-bit adder simulation."""
    caplog.set_level(logging.INFO, logger="deltacycle")

    period = 10

    # Inputs
    clk = Bool(name="clk")

    a = Bool(name="a")
    b = Bool(name="b")
    ci = Bool(name="ci")

    # Outputs
    s = Bool(name="s")
    co = Bool(name="co")

    async def drv_clk():
        clk.next = False
        while True:
            await sleep(period // 2)
            clk.next = not clk.prev

    async def drv_inputs():
        for a_val, b_val, ci_val, _, _ in VALS:
            a.next = a_val
            b.next = b_val
            ci.next = ci_val
            await clk.posedge()

    async def drv_outputs():
        vps = {a: a.changed, b: b.changed, ci: ci.changed}
        while True:
            await any_var(vps)
            g = a.value & b.value
            p = a.value | b.value
            s.next = a.value ^ b.value ^ ci.value
            co.next = g | p & ci.value

    async def mon_outputs():
        while True:
            await clk.posedge()
            logger.info("s=%d co=%d", s.prev, co.prev)

    async def main():
        create_task(drv_clk(), priority=0, name="drv_clk")
        create_task(drv_inputs(), priority=0, name="drv_inputs")
        create_task(drv_outputs(), priority=-1, name="drv_outputs")
        create_task(mon_outputs(), priority=1, name="mon_outputs")

    until = period * len(VALS)
    run(main(), until=until)

    # Check log messages
    msgs = [(r.time, r.getMessage()) for r in caplog.records]
    assert msgs == [
        (period * i + 5, f"s={s_val:d} co={co_val:d}")
        for i, (_, _, _, s_val, co_val) in enumerate(VALS)
    ]



"""Queue synchronization primitive."""

from collections import deque
from functools import cached_property

from ._kernel_if import KernelIf
from ._task import Task, WaitFifo


class Queue[T](KernelIf):
    """First-in, First-out (FIFO) queue."""

    def __init__(self, maxlen: int = 0):
        self._maxlen = maxlen
        self._items: deque[T] = deque()
        self._wait_not_empty = WaitFifo()
        self._wait_not_full = WaitFifo()

    def __len__(self) -> int:
        return len(self._items)

    @cached_property
    def _has_maxlen(self) -> bool:
        return self._maxlen > 0

    def empty(self) -> bool:
        """Return True if the queue is empty."""
        return not self._items

    def full(self) -> bool:
        """Return True if the queue is full."""
        return self._has_maxlen and len(self._items) == self._maxlen

    def _put(self, item: T):
        self._items.append(item)
        if self._wait_not_empty:
            task = self._wait_not_empty.pop()
            self._kernel.call_soon(task, value=(Task.Command.RESUME,))

    def try_put(self, item: T) -> bool:
        """Nonblocking put: Return True if a put attempt is successful."""
        if self.full():
            return False

        self._put(item)
        return True

    async def put(self, item: T):
        """Block until there is space to put the item."""
        if self.full():
            task = self._kernel.task()
            self._wait_not_full.push(task)
            y = await self._kernel.switch_coro()
            assert y is None

        self._put(item)

    def _get(self) -> T:
        item = self._items.popleft()
        if self._wait_not_full:
            task = self._wait_not_full.pop()
            self._kernel.call_soon(task, value=(Task.Command.RESUME,))
        return item

    def try_get(self) -> tuple[bool, T | None]:
        """Nonblocking get: Return True if a get attempt is successful."""
        if self.empty():
            return False, None

        item = self._get()
        return True, item

    async def get(self) -> T:
        """Block until an item is available to get."""
        if self.empty():
            task = self._kernel.task()
            self._wait_not_empty.push(task)
            y = await self._kernel.switch_coro()
            assert y is None

        return self._get()



"""Test seqlogic.sim module."""

import logging

from pytest import LogCaptureFixture

from deltacycle import Kernel, create_task, get_running_kernel, irun, run, sleep

from .common import Bool

logger = logging.getLogger("deltacycle")


async def drv_clk(clk: Bool):
    while True:
        await sleep(5)
        clk.next = not clk.prev


async def drv_a(a: Bool, clk: Bool):
    while True:
        await clk.edge()
        a.next = not a.prev


async def drv_b(b: Bool, clk: Bool):
    i = 0
    while True:
        await clk.edge()
        if i % 2 == 0:
            b.next = not b.prev
        else:
            b.next = b.prev
        i += 1


async def drv_c(c: Bool, clk: Bool):
    i = 0
    while True:
        await clk.edge()
        if i % 3 == 0:
            c.next = not c.prev
        i += 1


async def mon(a: Bool, b: Bool, c: Bool, clk: Bool):
    while True:
        await clk.edge()
        logger.info("a=%d b=%d c=%d", a.prev, b.prev, c.prev)


EXP = {
    (5, "a=0 b=0 c=0"),
    (10, "a=1 b=1 c=1"),
    (15, "a=0 b=1 c=1"),
    (20, "a=1 b=0 c=1"),
    (25, "a=0 b=0 c=0"),
    (30, "a=1 b=1 c=0"),
    (35, "a=0 b=1 c=0"),
    (40, "a=1 b=0 c=1"),
    (45, "a=0 b=0 c=1"),
}


def test_vars_run(caplog: LogCaptureFixture):
    """Test run, halt, run."""
    caplog.set_level(logging.INFO, logger="deltacycle")

    clk = Bool(name="clk")
    a = Bool(name="a")
    b = Bool(name="b")
    c = Bool(name="c")

    async def main():
        create_task(drv_clk(clk), priority=2)
        create_task(drv_a(a, clk), priority=2)
        create_task(drv_b(b, clk), priority=2)
        create_task(drv_c(c, clk), priority=2)
        create_task(mon(a, b, c, clk), priority=3)

    # Relative run limit
    run(main(), ticks=25)

    kernel = get_running_kernel()

    # Absolute run limit
    run(kernel=kernel, until=50)

    msgs = {(r.time, r.getMessage()) for r in caplog.records}
    assert msgs == EXP


def test_vars_iter(caplog: LogCaptureFixture):
    """Test iter, iter."""
    caplog.set_level(logging.INFO, logger="deltacycle")

    clk = Bool(name="clk")
    a = Bool(name="a")
    b = Bool(name="b")
    c = Bool(name="c")

    async def main():
        create_task(drv_clk(clk), priority=2)
        create_task(drv_a(a, clk), priority=2)
        create_task(drv_b(b, clk), priority=2)
        create_task(drv_c(c, clk), priority=2)
        create_task(mon(a, b, c, clk), priority=3)

    for t in irun(main()):
        if t >= 25:
            break

    kernel = get_running_kernel()
    assert kernel.state() is Kernel.State.RUNNING

    for t in irun(kernel=kernel):
        if t >= 50:
            break

    assert kernel.state() is Kernel.State.RUNNING

    msgs = {(r.time, r.getMessage()) for r in caplog.records}
    assert msgs == EXP


def test_vars_run_iter(caplog: LogCaptureFixture):
    """Test run, halt, iter."""
    caplog.set_level(logging.INFO, logger="deltacycle")

    clk = Bool(name="clk")
    a = Bool(name="a")
    b = Bool(name="b")
    c = Bool(name="c")

    async def main():
        create_task(drv_clk(clk), priority=2)
        create_task(drv_a(a, clk), priority=2)
        create_task(drv_b(b, clk), priority=2)
        create_task(drv_c(c, clk), priority=2)
        create_task(mon(a, b, c, clk), priority=3)

    # Relative run limit
    run(main(), ticks=25)

    kernel = get_running_kernel()

    for t in irun(kernel=kernel):
        if t >= 50:
            break

    assert kernel.state() is Kernel.State.RUNNING

    msgs = {(r.time, r.getMessage()) for r in caplog.records}
    assert msgs == EXP



"""Test deltacycle.TaskGroup"""

import logging

import pytest
from pytest import LogCaptureFixture

from deltacycle import Task, TaskGroup, get_current_task, run, sleep

logger = logging.getLogger("deltacycle")


async def cf_r(t: int, r: int):
    logger.info("enter")
    await sleep(t)
    logger.info("exit")
    return r


async def cf_x(t: int, r: int):
    logger.info("enter")
    await sleep(t)
    raise ArithmeticError(r)


async def cf_c(name: str, t0: int, r0: int, t1: int, r1: int):
    task = get_current_task()
    assert task.group is not None
    logger.info("enter")
    await sleep(t0)
    task.group.create_task(cf_r(t1, r1), name=name)
    logger.info("exit")
    return r0


EXP1 = {
    # Main
    (0, "main", "enter"),
    (15, "main", "exit"),
    # Coro 0
    (0, "C0", "enter"),
    (5, "C0", "exit"),
    # Coro 1
    (0, "C1", "enter"),
    (10, "C1", "exit"),
    # Coro 2
    (0, "C2", "enter"),
    (10, "C2", "exit"),
    # Coro 3
    (0, "C3", "enter"),
    (15, "C3", "exit"),
}


def test_group(caplog: LogCaptureFixture):
    caplog.set_level(logging.INFO, logger="deltacycle")

    async def main():
        logger.info("enter")

        ts: list[Task] = []
        async with TaskGroup() as tg:
            ts.append(tg.create_task(cf_r(5, 0), name="C0"))
            ts.append(tg.create_task(cf_r(10, 1), name="C1"))
            ts.append(tg.create_task(cf_r(10, 2), name="C2"))
            ts.append(tg.create_task(cf_r(15, 3), name="C3"))

        logger.info("exit")

        assert ts[0].result() == 0
        assert ts[1].result() == 1
        assert ts[2].result() == 2
        assert ts[3].result() == 3

    run(main())
    msgs = {(r.time, r.taskName, r.getMessage()) for r in caplog.records}
    assert msgs == EXP1


EXP2 = {
    # Main
    (0, "main", "enter"),
    (10, "main", "exit"),
    # Weirdo
    (0, "X0", "enter"),
    # Coro 0 - completes
    (0, "C0", "enter"),
    (5, "C0", "exit"),
    # Coro 1 - completes
    (0, "C1", "enter"),
    (10, "C1", "exit"),
    # Coro 2 - raises exception
    (0, "C2", "enter"),
    # Coro 3 - raises exception
    (0, "C3", "enter"),
    # Coro 4 - completes
    (0, "C4", "enter"),
    (10, "C4", "exit"),
    # Coro 5,6,7 - interrupted
    (0, "C5", "enter"),
    (0, "C6", "enter"),
    (0, "C7", "enter"),
}


def test_group_child_except(caplog: LogCaptureFixture):
    """One child raises an exception, others are interrupted."""
    caplog.set_level(logging.INFO, logger="deltacycle")

    async def main():
        logger.info("enter")

        ts: list[Task] = []
        with pytest.raises(ExceptionGroup) as e:
            async with TaskGroup() as tg:
                # Handle weird case of done child
                await tg.create_task(sleep(0))

                # Another weird case of done child that raised an exception
                try:
                    await tg.create_task(cf_x(0, 69), name="X0")
                except ArithmeticError:
                    pass

                # These tasks will complete successfully
                ts.append(tg.create_task(cf_r(5, 0), name="C0"))
                ts.append(tg.create_task(cf_r(10, 1), name="C1"))
                # These tasks will raise an exception
                ts.append(tg.create_task(cf_x(10, 2), name="C2"))
                ts.append(tg.create_task(cf_x(10, 3), name="C3"))
                # This task will also complete successfully
                # (It completes before interrupt takes effect)
                ts.append(tg.create_task(cf_r(10, 4), name="C4"))
                # These tasks will be interrupted
                ts.append(tg.create_task(cf_r(11, 5), name="C5"))
                ts.append(tg.create_task(cf_r(13, 6), name="C6"))
                ts.append(tg.create_task(cf_r(15, 7), name="C7"))

        assert ts[0].result() == 0
        assert ts[1].result() == 1
        assert ts[4].result() == 4

        assert ts[5].state() is Task.State.EXCEPTED
        assert ts[6].state() is Task.State.EXCEPTED
        assert ts[7].state() is Task.State.EXCEPTED

        exc = ts[2].exception()
        assert isinstance(exc, ArithmeticError) and exc.args == (2,)
        exc = ts[3].exception()
        assert isinstance(exc, ArithmeticError) and exc.args == (3,)

        excs = e.value.args[1]
        assert [type(exc) for exc in excs] == [ArithmeticError, ArithmeticError]
        assert [exc.args for exc in excs] == [(2,), (3,)]

        logger.info("exit")

    run(main())

    msgs = {(r.time, r.taskName, r.getMessage()) for r in caplog.records}
    assert msgs == EXP2


EXP3 = {
    (0, "main", "enter"),
    (0, "main", "exit"),
}


def test_group_except(caplog: LogCaptureFixture):
    caplog.set_level(logging.INFO, logger="deltacycle")

    async def main():
        logger.info("enter")

        ts: list[Task] = []
        with pytest.raises(ArithmeticError) as e:
            async with TaskGroup() as tg:
                # Handle weird case of done child
                await tg.create_task(sleep(0))

                ts.append(tg.create_task(cf_r(5, 0), name="C0"))
                ts.append(tg.create_task(cf_r(10, 1), name="C1"))

                raise ArithmeticError(42)

        assert ts[0].state() is Task.State.EXCEPTED
        assert ts[1].state() is Task.State.EXCEPTED
        assert e.value.args == (42,)

        logger.info("exit")

    run(main())
    msgs = {(r.time, r.taskName, r.getMessage()) for r in caplog.records}
    assert msgs == EXP3


EXP4 = {
    # Main
    (0, "main", "enter"),
    (10, "main", "exit"),
    # Coros 0,1,2,3
    (0, "C0", "enter"),
    (2, "C0", "exit"),
    (0, "C1", "enter"),
    (2, "C1", "exit"),
    (0, "C2", "enter"),
    (2, "C2", "exit"),
    (0, "C3", "enter"),
    (2, "C3", "exit"),
    # Newborns 0,1 - completes
    (2, "N0", "enter"),
    (9, "N0", "exit"),
    (2, "N1", "enter"),
    (10, "N1", "exit"),
    # Newborn 2 - interrupted
    (2, "N2", "enter"),
    (2, "N3", "enter"),
    # Coro 4,5 - completes
    (0, "C4", "enter"),
    (5, "C4", "exit"),
    (0, "C5", "enter"),
    (10, "C5", "exit"),
    # Coro 6,7 - raises exception
    (0, "C6", "enter"),
    (0, "C7", "enter"),
    # Coro 8 - completes
    (10, "C8", "exit"),
    (0, "C8", "enter"),
    # Coros 9,10 - interrupted
    (0, "C9", "enter"),
    (0, "C10", "enter"),
}


def test_group_newborns_except(caplog: LogCaptureFixture):
    caplog.set_level(logging.INFO, logger="deltacycle")

    async def main():
        logger.info("enter")

        ts: list[Task] = []
        with pytest.raises(ExceptionGroup) as e:
            async with TaskGroup() as tg:
                # Newborns
                ts.append(tg.create_task(cf_c("N0", 2, 0, 7, 10), name="C0"))
                ts.append(tg.create_task(cf_c("N1", 2, 1, 8, 11), name="C1"))
                ts.append(tg.create_task(cf_c("N2", 2, 2, 9, 12), name="C2"))
                ts.append(tg.create_task(cf_c("N3", 2, 3, 10, 13), name="C3"))

                # These tasks will complete successfully
                ts.append(tg.create_task(cf_r(5, 4), name="C4"))
                ts.append(tg.create_task(cf_r(10, 5), name="C5"))
                # These tasks will raise an exception
                ts.append(tg.create_task(cf_x(10, 6), name="C6"))
                ts.append(tg.create_task(cf_x(10, 7), name="C7"))
                # This task will also complete successfully
                # (It completes before interrupt takes effect)
                ts.append(tg.create_task(cf_r(10, 8), name="C8"))
                # These tasks will be interrupted
                ts.append(tg.create_task(cf_r(11, 9), name="C9"))
                ts.append(tg.create_task(cf_r(12, 10), name="C10"))

        assert ts[4].result() == 4
        assert ts[5].result() == 5
        assert ts[8].result() == 8

        assert ts[9].state() is Task.State.EXCEPTED
        assert ts[10].state() is Task.State.EXCEPTED

        exc = ts[6].exception()
        assert isinstance(exc, ArithmeticError) and exc.args == (6,)
        exc = ts[7].exception()
        assert isinstance(exc, ArithmeticError) and exc.args == (7,)

        excs = e.value.args[1]
        assert [type(exc) for exc in excs] == [ArithmeticError, ArithmeticError]
        assert [exc.args for exc in excs] == [(6,), (7,)]

        logger.info("exit")

    run(main())

    msgs = {(r.time, r.taskName, r.getMessage()) for r in caplog.records}
    assert msgs == EXP4



"""Hello, world!"""

import logging

import pytest
from pytest import LogCaptureFixture

from deltacycle import Kernel, get_kernel, run, sleep

logger = logging.getLogger("deltacycle")


EXP = [
    (-1, "Before Time"),
    (2, "Hello"),
    (4, "World"),
]


def test_hello(caplog: LogCaptureFixture):
    """Test basic async/await hello world functionality."""
    caplog.set_level(logging.INFO, logger="deltacycle")

    logger.info("Before Time")

    async def hello():
        await sleep(2)
        logger.info("Hello")
        await sleep(2)
        logger.info("World")
        return 42

    ret = run(hello())
    assert ret == 42

    kernel = get_kernel()
    assert kernel is not None
    assert kernel.state() is Kernel.State.COMPLETED

    with pytest.raises(RuntimeError):
        run(kernel=kernel)

    msgs = [(r.time, r.getMessage()) for r in caplog.records]
    assert msgs == EXP



"""Event synchronization primitive"""

from __future__ import annotations

from collections.abc import Generator

from ._kernel_if import KernelIf
from ._task import Task, WaitFifo


class Event(KernelIf):
    """Notify multiple tasks that some event has happened."""

    def __init__(self):
        self._flag = False
        self._waiting = WaitFifo()

    def __bool__(self) -> bool:
        return self._flag

    def __await__(self) -> Generator[None, Event, Event]:
        if not self._flag:
            task = self._kernel.task()
            self._wait(task)
            e = yield from self._kernel.switch_gen()
            assert e is self

        return self

    def __or__(self, other: Event) -> EventList:
        return EventList(self, other)

    def _wait(self, task: Task):
        self._waiting.push(task)
        self._kernel._task2events[task].add(self)

    def _set(self):
        while self._waiting:
            task = self._waiting.pop()

            # Remove task from Event waiting queues
            self._kernel._task2events[task].remove(self)
            while self._kernel._task2events[task]:
                e = self._kernel._task2events[task].pop()
                e._waiting.drop(task)
            del self._kernel._task2events[task]

            # Send event id to parent task
            self._kernel.call_soon(task, value=(Task.Command.RESUME, self))

    def set(self):
        self._flag = True
        self._set()

    def clear(self):
        self._flag = False


class EventList(KernelIf):
    def __init__(self, *events: Event):
        self._events = events

    def __await__(self) -> Generator[None, Event, Event]:
        task = self._kernel.task()

        fst = None
        for e in self._events:
            if e:
                fst = e
                break

        # No events set yet
        if fst is None:
            # Await first event to be set
            for e in self._events:
                e._wait(task)
            fst = yield from self._kernel.switch_gen()
            assert isinstance(fst, Event)

        return fst

    def __or__(self, other: Event) -> EventList:
        return EventList(*self._events, other)



"""Test variables"""

import logging

from deltacycle import Singular, create_task, run, sleep

logger = logging.getLogger("deltacycle")


def test_var_await():
    x = Singular(value=0)
    x2 = Singular(value=0)
    x4 = Singular(value=0)
    x8 = Singular(value=0)

    async def cf(y: Singular, x: Singular):
        while True:
            _ = await x
            y.next = 2 * x.value

    async def main():
        create_task(cf(x2, x))
        create_task(cf(x4, x2))
        create_task(cf(x8, x4))

        for i in range(100):
            await sleep(1)
            x.next = i
            await sleep(1)
            assert x2.value == 2 * i
            assert x4.value == 4 * i
            assert x8.value == 8 * i

    run(main())



"""Model variables"""

from __future__ import annotations

from abc import ABC
from collections import defaultdict
from collections.abc import Generator, Hashable

from ._kernel_if import KernelIf
from ._task import Predicate, Task, WaitSet


class Variable(KernelIf):
    """Model component.

    Children::

               Variable
                  |
           +------+------+
           |             |
        Singular     Aggregate
    """

    def __init__(self):
        self._waiting = WaitSet()

    def __await__(self) -> Generator[None, Variable, Variable]:
        task = self._kernel.task()
        self._wait(self.changed, task)
        v = yield from self._kernel.switch_gen()
        assert v is self
        return self

    def _wait(self, p: Predicate, task: Task):
        self._waiting.push((p, task))
        self._kernel._task2vars[task].add(self)

    def _set(self):
        self._waiting.set()

        while self._waiting:
            task = self._waiting.pop()

            # Remove task from Variable waiting queues
            self._kernel._task2vars[task].remove(self)
            while self._kernel._task2vars[task]:
                v = self._kernel._task2vars[task].pop()
                v._waiting.drop(task)
            del self._kernel._task2vars[task]

            # Send variable id to parent task
            self._kernel.call_soon(task, value=(Task.Command.RESUME, self))

        # Add variable to update set
        self._kernel.touch(self)

    def changed(self) -> bool:
        """Return True if changed during the current time slot."""
        raise NotImplementedError()  # pragma: no cover

    def update(self) -> None:
        """Kernel callback."""
        raise NotImplementedError()  # pragma: no cover


class Value[T](ABC):
    """Variable value."""

    def get_prev(self) -> T:
        raise NotImplementedError()  # pragma: no cover

    prev = property(fget=get_prev)

    def set_next(self, value: T) -> None:
        raise NotImplementedError()  # pragma: no cover

    next = property(fset=set_next)


class Singular[T](Variable, Value[T]):
    """Model state organized as a single unit."""

    def __init__(self, value: T):
        Variable.__init__(self)
        self._prev = value
        self._next = value
        self._changed: bool = False

    # Value
    def get_prev(self) -> T:
        return self._prev

    prev = property(fget=get_prev)

    def set_next(self, value: T):
        self._changed = value != self._next
        self._next = value

        # Notify the kernel
        self._set()

    next = property(fset=set_next)

    # Variable
    def get_value(self) -> T:
        return self._next

    value = property(fget=get_value)

    def changed(self) -> bool:
        return self._changed

    def update(self):
        self._prev = self._next
        self._changed = False


class Aggregate[T](Variable):
    """Model state organized as multiple units."""

    def __init__(self, value: T):
        Variable.__init__(self)
        self._prevs: dict[Hashable, T] = defaultdict(lambda: value)
        self._nexts: dict[Hashable, T] = dict()

    # [key] => Value
    def __getitem__(self, key: Hashable) -> AggrItem[T]:
        return AggrItem(self, key)

    def get_prev(self, key: Hashable) -> T:
        return self._prevs[key]

    def get_next(self, key: Hashable) -> T:
        try:
            return self._nexts[key]
        except KeyError:
            return self._prevs[key]

    def set_next(self, key: Hashable, value: T):
        if value != self.get_next(key):
            self._nexts[key] = value

        # Notify the kernel
        self._set()

    # Variable
    def get_value(self) -> AggrValue[T]:
        return AggrValue(self)

    value = property(fget=get_value)

    def changed(self) -> bool:
        return bool(self._nexts)

    def update(self):
        while self._nexts:
            key, value = self._nexts.popitem()
            self._prevs[key] = value


class AggrItem[T](Value[T]):
    """Wrap Aggregate __getitem__."""

    def __init__(self, aggr: Aggregate[T], key: Hashable):
        self._aggr = aggr
        self._key = key

    def get_prev(self) -> T:
        return self._aggr.get_prev(self._key)

    prev = property(fget=get_prev)

    def set_next(self, value: T):
        self._aggr.set_next(self._key, value)

    next = property(fset=set_next)


class AggrValue[T]:
    """Wrap Aggregate value."""

    def __init__(self, aggr: Aggregate[T]):
        self._aggr = aggr

    def __getitem__(self, key: Hashable) -> T:
        return self._aggr.get_next(key)



"""Test seqlogic.sim.Lock class."""

import logging

from pytest import LogCaptureFixture

from deltacycle import Lock, create_task, run, sleep

logger = logging.getLogger("deltacycle")


async def use_acquire_release(lock: Lock, t1: int, t2: int):
    logger.info("enter")

    await sleep(t1)

    logger.info("attempt get")
    await lock.get()
    logger.info("acquired")

    try:
        await sleep(t2)
    finally:
        logger.info("put")
        lock.put()

    await sleep(10)
    logger.info("exit")


async def use_with(lock: Lock, t1: int, t2: int):
    logger.info("enter")

    await sleep(t1)

    logger.info("attempt get")
    async with lock:
        logger.info("acquired")
        await sleep(t2)
    logger.info("put")

    await sleep(10)
    logger.info("exit")


EXP = {
    (0, "0", "enter"),
    (0, "1", "enter"),
    (0, "2", "enter"),
    (0, "3", "enter"),
    (10, "0", "attempt get"),
    (10, "0", "acquired"),
    (11, "1", "attempt get"),
    (12, "2", "attempt get"),
    (13, "3", "attempt get"),
    (20, "0", "put"),
    (20, "1", "acquired"),
    (30, "0", "exit"),
    (30, "1", "put"),
    (30, "2", "acquired"),
    (40, "1", "exit"),
    (40, "2", "put"),
    (40, "3", "acquired"),
    (50, "2", "exit"),
    (50, "3", "put"),
    (60, "3", "exit"),
}


def test_acquire_release(caplog: LogCaptureFixture):
    caplog.set_level(logging.INFO, logger="deltacycle")

    async def main():
        lock = Lock()
        for i in range(4):
            create_task(use_acquire_release(lock, i + 10, 10), name=f"{i}")

    run(main())

    msgs = {(r.time, r.taskName, r.getMessage()) for r in caplog.records}
    assert msgs == EXP


def test_async_with(caplog: LogCaptureFixture):
    caplog.set_level(logging.INFO, logger="deltacycle")

    async def main():
        lock = Lock()
        for i in range(4):
            create_task(use_with(lock, i + 10, 10), name=f"{i}")

    run(main())

    msgs = {(r.time, r.taskName, r.getMessage()) for r in caplog.records}
    assert msgs == EXP



"""Simulate a register file."""

import logging

from pytest import LogCaptureFixture

from deltacycle import any_var, create_task, run, sleep

from .common import Bool, Int, IntMem

logger = logging.getLogger("deltacycle")


# wr_en, wr_addr, wr_data, rd_addr, rd_data
VALS = [
    (True, 0, 42, 0, 0),
    (True, 1, 43, 0, 42),
    (True, 2, 44, 1, 43),
    (True, 3, 45, 2, 44),
    (True, 4, 46, 3, 45),
    (True, 5, 47, 4, 46),
    (True, 6, 48, 5, 47),
    (True, 7, 49, 6, 48),
    (True, 8, 50, 7, 49),
    (True, 9, 51, 8, 50),
    # Note: Same WrData
    (True, 0, 42, 9, 51),
    (False, 0, 0, 0, 42),
]


def test_regfile(caplog: LogCaptureFixture):
    caplog.set_level(logging.INFO, logger="deltacycle")

    clk = Bool(name="clk")
    period = 10

    wr_en = Bool(name="wr_en")
    wr_addr = Int(name="wr_addr")
    wr_data = Int(name="wr_data")

    rd_addr = Int(name="rd_addr")
    rd_data = Int(name="rd_data")

    # State
    regs = IntMem(name="regs")

    async def drv_clk():
        clk.next = False
        while True:
            await sleep(period // 2)
            clk.next = not clk.prev

    async def drv_inputs():
        for we, wa, wd, ra, _ in VALS:
            wr_en.next = we
            wr_addr.next = wa
            wr_data.next = wd
            rd_addr.next = ra
            await clk.posedge()

    async def mon_outputs():
        while True:
            await clk.posedge()
            logger.info(
                "wr_en=%d wr_addr=%d wr_data=%d rd_addr=%d rd_data=%d",
                wr_en.prev,
                wr_addr.prev,
                wr_data.prev,
                rd_addr.prev,
                rd_data.prev,
            )

    async def wr_port():
        def clk_pred():
            return clk.is_posedge() and wr_en.prev

        while True:
            await any_var({clk: clk_pred})
            regs[wr_addr.prev].next = wr_data.prev

    async def rd_port():
        vps = {regs: regs.changed, rd_addr: rd_addr.changed}
        while True:
            await any_var(vps)
            rd_data.next = regs.value[rd_addr.value]

    async def main():
        create_task(drv_clk(), priority=2)
        create_task(drv_inputs(), priority=2)
        create_task(mon_outputs(), priority=3)
        create_task(wr_port(), priority=2)
        create_task(rd_port(), priority=1)

    run(main(), until=120)

    # Check log messages
    msgs = [(r.time, r.getMessage()) for r in caplog.records]
    assert msgs == [
        (period * i + 5, f"wr_en={we:d} wr_addr={wa} wr_data={wd} rd_addr={ra} rd_data={rd}")
        for i, (we, wa, wd, ra, rd) in enumerate(VALS)
    ]

    for i in range(10):
        assert regs[i].prev == 42 + i



"""KernelIf abstract base class

Allows easy access to global kernel for Event, Semaphore, Task, ...
Works around tricky circular import: Kernel => Task => Kernel.
"""

from abc import ABC
from functools import cached_property


class KernelIf(ABC):
    @cached_property
    def _kernel(self):
        from ._top import get_running_kernel  # noqa: PLC0415

        return get_running_kernel()
